import numpy as np

import imglyb
from jnius import autoclass, MetaJavaClass

NativeType = autoclass('net.imglib2.type.NativeType')

FloatType                 = autoclass('net.imglib2.type.numeric.real.FloatType')
DoubleType                = autoclass('net.imglib2.type.numeric.real.DoubleType')
ByteType                  = autoclass('net.imglib2.type.numeric.integer.ByteType')
UnsignedByteType          = autoclass('net.imglib2.type.numeric.integer.UnsignedByteType')
ShortType                 = autoclass('net.imglib2.type.numeric.integer.ShortType')
UnsignedShortType         = autoclass('net.imglib2.type.numeric.integer.UnsignedShortType')
IntType                   = autoclass('net.imglib2.type.numeric.integer.IntType')
UnsignedIntType           = autoclass('net.imglib2.type.numeric.integer.UnsignedIntType')
LongType                  = autoclass('net.imglib2.type.numeric.integer.LongType')
UnsignedLongType          = autoclass('net.imglib2.type.numeric.integer.UnsignedLongType')

VolatileFloatType         = autoclass('net.imglib2.type.volatiles.VolatileFloatType')
VolatileDoubleType        = autoclass('net.imglib2.type.volatiles.VolatileDoubleType')
VolatileByteType          = autoclass('net.imglib2.type.volatiles.VolatileByteType')
VolatileUnsignedByteType  = autoclass('net.imglib2.type.volatiles.VolatileUnsignedByteType')
VolatileShortType         = autoclass('net.imglib2.type.volatiles.VolatileShortType')
VolatileUnsignedShortType = autoclass('net.imglib2.type.volatiles.VolatileUnsignedShortType')
VolatileIntType           = autoclass('net.imglib2.type.volatiles.VolatileIntType')
VolatileUnsignedIntType   = autoclass('net.imglib2.type.volatiles.VolatileUnsignedIntType')
VolatileLongType          = autoclass('net.imglib2.type.volatiles.VolatileLongType')
VolatileUnsignedLongType  = autoclass('net.imglib2.type.volatiles.VolatileUnsignedLongType')


def for_np_dtype(dtype, volatile=False):
    if dtype == np.uint8:
        return VolatileUnsignedByteType() if volatile else UnsignedByteType()
    if dtype == np.int8:
        return VolatileByteType() if volatile else ByteType()

    if dtype == np.uint16:
        return VolatileUnsignedShortType() if volatile else UnsignedShortType()
    if dtype == np.int16:
        return VolatileShortType() if volatile else ShortType()

    if dtype == np.uint32:
        return VolatileUnsignedIntType() if volatile else UnsignedIntType()
    if dtype == np.int32:
        return VolatileIntType() if volatile else IntType()

    if dtype == np.uint64:
        return VolatileUnsignedLongType() if volatile else UnsignedLongType()
    if dtype == np.int64:
        return VolatileLongType() if volatile else LongType()

    if dtype == np.float32:
        return VolatileFloatType() if volatile else FloatType()
    if dtype == np.float64:
        return VolatileDoubleType() if volatile else DoubleType()




from imglyb import util
from jnius import autoclass, cast

import ctypes
import numpy as np

Intervals = autoclass('net.imglib2.util.Intervals')

dtype_selector = {
    'FloatType': np.dtype('float32'),
    'DoubleType': np.dtype('float64'),
    'ByteType': np.dtype('int8'),
    'UnsignedByteType': np.dtype('uint8'),
    'ShortType': np.dtype('int16'),
    'UnsignedShortType': np.dtype('uint16'),
    'IntType': np.dtype('int32'),
    'UnsignedIntType': np.dtype('uint32'),
    'LongType': np.dtype('int64'),
    'UnsignedLongType': np.dtype('uint64'),
}

ctype_conversions_imglib = {
    'FloatType': ctypes.c_float,
    'DoubleType': ctypes.c_double,
    'ByteType': ctypes.c_int8,
    'UnsignedByteType': ctypes.c_uint8,
    'ShortType': ctypes.c_int16,
    'UnsignedShortType': ctypes.c_uint16,
    'IntType': ctypes.c_int32,
    'UnsignedIntType': ctypes.c_uint32,
    'LongType': ctypes.c_int64,
    'UnsignedLongType': ctypes.c_uint64
}


def get_address(rai):
    class_name = util.Helpers.classNameSimple(rai)
    class_name_full = util.Helpers.className(rai)
    if class_name in ('ArrayImg', 'UnsafeImg'):
        access = cast(class_name_full, rai).update(None)
        access_type = util.Helpers.className(access)

        if 'basictypelongaccess.unsafe' in access_type:
            return cast(access_type, access).getAddress()
        else:
            raise ValueError("Excpected unsafe access but got {}".format(access_type))

    else:
        raise ValueError("Excpected ArrayImg or UnsafeImg but got {}".format(class_name))


class ImgLibReferenceGuard(np.ndarray):
    def __new__(cls, rai):
        access = rai.randomAccess()
        rai.min(access)
        imglib_type = util.Helpers.classNameSimple(access.get())
        address = get_address(rai)

        shape = tuple(Intervals.dimensionsAsLongArray(rai))[::-1]
        dtype = dtype_selector[imglib_type]
        pointer = ctypes.cast(address, ctypes.POINTER(ctype_conversions_imglib[imglib_type]))
        order = 'C'

        obj = np.ndarray.__new__(cls, buffer=np.ctypeslib.as_array(pointer, shape=shape), shape=shape, dtype=dtype)
        obj.setflags(write=True)
        obj.rai = rai
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.rai = obj.rai


if __name__ == "__main__":

    ArrayImgs = autoclass('net.imglib2.img.array.ArrayImgs')
    UnsafeUtil = autoclass('net.imglib2.img.basictypelongaccess.unsafe.UnsafeUtil')
    Arrays = autoclass('java.util.Arrays')
    OwningFloatUnsafe = autoclass('net.imglib2.img.basictypelongaccess.unsafe.owning.OwningFloatUnsafe')
    Fraction = autoclass('net.imglib2.util.Fraction')
    LongStream = autoclass('java.util.stream.LongStream')

    shape = (2, 3, 4)
    n_elements = int(np.prod(shape))
    data_store = OwningFloatUnsafe(n_elements)
    dim_array = LongStream.of(*shape).toArray()
    print(Arrays.toString(dim_array))
    rai = util.Helpers.toArrayImg(cast(util.Helpers.className(data_store), data_store), dim_array)
    # rai = ArrayImgs.floats( *shape )
    c = rai.cursor()
    count = 23
    while c.hasNext():
        c.next().setReal(count)
        count += 1
    print(util.Helpers.className(rai.randomAccess().get()))
    print(util.Helpers.classNameSimple(rai.randomAccess().get()))
    arr = ImgLibReferenceGuard(rai)
    print(arr, arr.mean())
    c = rai.cursor()
    c.fwd()
    c.next().setReal(0)
    print(arr, arr.mean())



from __future__ import print_function

import logging
import sys

_logger = logging.getLogger(__name__)

__all__ = ('to_imglib', 'to_imglib_argb', 'to_numpy')

def _init_jvm_options():
    import imglyb_config
    import jnius_config
    import scyjava_config

    imglib2_imglyb_version = imglyb_config.get_imglib2_imglyb_version()

    if jnius_config.vm_running:
        _logger.warning('JVM is already running, will not add relevant endpoints to classpath -- '
                        'required classes might not be on classpath. '
                        'In case of failure, try importing imglyb before scyjava or jnius')
        return jnius_config

    IMGLIB2_IMGLYB_ENDPOINT = 'net.imglib2:imglib2-imglyb:{}'.format(imglib2_imglyb_version)
    RELEVANT_MAVEN_REPOS    = {
        'scijava.public' : scyjava_config.maven_scijava_repository()
    }
    for _, repo in scyjava_config.get_repositories().items():
        if 'scijava.public' in RELEVANT_MAVEN_REPOS and repo == RELEVANT_MAVEN_REPOS['scijava.public']:
            del RELEVANT_MAVEN_REPOS['scijava.public']
    scyjava_config.add_repositories(RELEVANT_MAVEN_REPOS)
    scyjava_config.add_endpoints(IMGLIB2_IMGLYB_ENDPOINT)

    import scyjava

    return jnius_config


config = _init_jvm_options()

from .imglib_ndarray import ImgLibReferenceGuard as _ImgLibReferenceGuard
from .ndarray_like_as_img import \
    as_cell_img, \
    as_cell_img_with_array_accesses, \
    as_cell_img_with_native_accesses
from .util import \
    to_imglib, \
    to_imglib_argb


def to_numpy(source):
    return _ImgLibReferenceGuard(source)



import os

__all__ = ('set_imglib2_imglyb_version', 'get_imglib2_imglyb_version')

__major__   = 0
__minor__   = 4
__patch__   = 0
__tag__     = ''
__version__ = f'{__major__}.{__minor__}.{__patch__}.{__tag__}'.strip('.')

class _Version(object):

    def major(self):
        return __major__

    def minor(self):
        return __minor__

    def patch(self):
        return __patch__

    def tag(self):
        return __tag__

    def version(self):
        return __version__

    def __str__(self):
        return self.version()

_version = _Version()

version = _version.version()

_default_imglib2_imglyb_version = '0.4.0'
_imglib2_imglyb_version         = os.getenv('IMGLIB2_IMGLYB_VERSION', _default_imglib2_imglyb_version)

def set_imglib2_imglyb_version(version):
    global _imglib2_imglyb_version
    _imglib2_imglyb_version = version

def get_imglib2_imglyb_version():
    global _imglib2_imglyb_version
    return _imglib2_imglyb_version



from __future__ import division

import logging
import numpy as np

from jnius import autoclass, PythonJavaClass, java_method

_logger = logging.getLogger(__name__)

__all__ = (
    'to_imglib',
    'to_imglib_argb',
    'options2D',
    'RunnableFromFunc'
)

# java
Random = autoclass('java.util.Random')

# imglib
Helpers                            = autoclass('net.imglib2.python.Helpers')
NumpyToImgLibConversions           = autoclass('net.imglib2.python.NumpyToImgLibConversions')
NumpyToImgLibConversionsWithStride = autoclass('net.imglib2.python.NumpyToImgLibConversionsWithStride')
Views                              = autoclass('net.imglib2.view.Views')

# bigdataviewer
BdvFunctions = autoclass('bdv.util.BdvFunctions')
BdvOptions   = autoclass('bdv.util.BdvOptions')

# Guard
ReferenceGuardingRandomAccessibleInterval = autoclass('net/imglib2/python/ReferenceGuardingRandomAccessibleInterval')

numpy_dtype_to_conversion_method = {
    np.dtype('complex64')  : NumpyToImgLibConversions.toComplexFloat,
    np.dtype('complex128') : NumpyToImgLibConversions.toComplexDouble,
    np.dtype('float32')    : NumpyToImgLibConversions.toFloat,
    np.dtype('float64')    : NumpyToImgLibConversions.toDouble,
    np.dtype('int8')       : NumpyToImgLibConversions.toByte,
    np.dtype('int16')      : NumpyToImgLibConversions.toShort,
    np.dtype('int32')      : NumpyToImgLibConversions.toInt,
    np.dtype('int64')      : NumpyToImgLibConversions.toLong,
    np.dtype('uint8')      : NumpyToImgLibConversions.toUnsignedByte,
    np.dtype('uint16')     : NumpyToImgLibConversions.toUnsignedShort,
    np.dtype('uint32')     : NumpyToImgLibConversions.toUnsignedInt,
    np.dtype('uint64')     : NumpyToImgLibConversions.toUnsignedLong
}

numpy_dtype_to_conversion_with_stride_method = {
    np.dtype('complex64')  : NumpyToImgLibConversionsWithStride.toComplexFloat,
    np.dtype('complex128') : NumpyToImgLibConversionsWithStride.toComplexDouble,
    np.dtype('float32')    : NumpyToImgLibConversionsWithStride.toFloat,
    np.dtype('float64')    : NumpyToImgLibConversionsWithStride.toDouble,
    np.dtype('int8')       : NumpyToImgLibConversionsWithStride.toByte,
    np.dtype('int16')      : NumpyToImgLibConversionsWithStride.toShort,
    np.dtype('int32')      : NumpyToImgLibConversionsWithStride.toInt,
    np.dtype('int64')      : NumpyToImgLibConversionsWithStride.toLong,
    np.dtype('uint8')      : NumpyToImgLibConversionsWithStride.toUnsignedByte,
    np.dtype('uint16')     : NumpyToImgLibConversionsWithStride.toUnsignedShort,
    np.dtype('uint32')     : NumpyToImgLibConversionsWithStride.toUnsignedInt,
    np.dtype('uint64')     : NumpyToImgLibConversionsWithStride.toUnsignedLong
}


def _get_address(source):
    return source.ctypes.data


class ReferenceGuard(PythonJavaClass):
    __javainterfaces__ = ['net/imglib2/python/ReferenceGuardingRandomAccessibleInterval$ReferenceHolder']

    def __init__(self, *args, **kwargs):
        super(ReferenceGuard, self).__init__()
        self.args = args


# how to use type hints for python < 3.5?
def to_imglib(source):
    return ReferenceGuardingRandomAccessibleInterval(_to_imglib(source), ReferenceGuard(source))


# how to use type hints for python < 3.5?
def _to_imglib(source):
    address = _get_address(source)
    if not source.dtype in numpy_dtype_to_conversion_method:
        raise NotImplementedError("Cannot convert dtype to ImgLib2 type yet: {}".format(source.dtype))
    elif source.flags['CARRAY']:
        return numpy_dtype_to_conversion_method[source.dtype](address, *source.shape[::-1])
    else:
        stride = np.array(source.strides[::-1]) / source.itemsize
        return numpy_dtype_to_conversion_with_stride_method[source.dtype](address, tuple(stride), source.shape[::-1])


def to_imglib_argb(source):
    return ReferenceGuardingRandomAccessibleInterval(_to_imglib_argb(source), ReferenceGuard(source))


def _to_imglib_argb(source):
    address = _get_address(source)
    if not (source.dtype == np.dtype('int32') or source.dtype == np.dtype('uint32')):
        raise NotImplementedError("source.dtype must be int32 or uint32")
    if source.flags['CARRAY']:
        return NumpyToImgLibConversions.toARGB(address, *source.shape[::-1])
    else:
        stride = np.array(source.strides[::-1]) / source.itemsize
        return NumpyToImgLibConversionsWithStride.toARGB(address, tuple(stride), source.shape[::-1])


def options2D():
    return BdvOptions.options().is2D()


class GenericOverlayRenderer(PythonJavaClass):
    __javainterfaces__ = ['net/imglib2/ui/OverlayRenderer']

    def __init__(self, draw_overlays=lambda g: None, set_canvas_size=lambda w, h: None):
        super(GenericOverlayRenderer, self).__init__()
        self.draw_overlays = draw_overlays
        self.set_canvas_size = set_canvas_size

    @java_method('(Ljava/awt/Graphics;)V')
    def drawOverlays(self, g):
        self.draw_overlays(g)

    @java_method('(II)V')
    def setCanvasSize(self, width, height):
        self.set_canvas_size(width, height)


class GenericMouseMotionListener(PythonJavaClass):
    __javainterfaces__ = ['java/awt/event/MouseMotionListener']

    def __init__(self, mouse_dragged=lambda e: None, mouse_moved=lambda e: None):
        super(GenericMouseMotionListener, self).__init__()
        self.mouse_dragged = mouse_dragged
        self.mouse_moved = mouse_moved

    @java_method('(Ljava/awt/event/MouseEvent;)V')
    def mouseDragged(self, e):
        self.mouse_dragged(e)

    @java_method('(Ljava/awt/event/MouseEvent;)V')
    def mouseMoved(self, e):
        self.mouse_moved(e)

class RunnableFromFunc(PythonJavaClass):
    __javainterfaces__ = ['java/lang/Runnable']

    def __init__(self, func):
        super(RunnableFromFunc, self).__init__()
        self.func = func

    @java_method('()V')
    def run(self):
        _logger.debug('Running function %s', self.func)
        self.func()



import threading

'''

'''
class ReferenceStore(object):

    def __init__(self):
        self.lock    = threading.RLock()
        self.store   = dict()
        self.next_id = 0

    def add_reference(self, ref_id, reference):
        with self.lock:
            self.store[ref_id] = reference

    def add_reference_with_new_id(self, reference):
        ref_id = self.get_next_id()
        self.add_reference(ref_id, reference)
        return ref_id

    def get_next_id(self):
        with self.lock:
            next_id = self.next_id
            self.next_id += 1
        return next_id

    def remove_reference(self, ref_id):
        with self.lock:
            if ref_id in self.store:
                del self.store[ref_id]

    def clear(self):
        with self.lock:
            self.store.clear()

    def reset_next_id(self):
        with self.lock:
            self.next_id = 0

    def number_of_stored_references(self):
        return len(self.store)
